/* $Id: isaac_utils.c,v 1.29 2013-03-12 08:06:48 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:06:48 $
 * $Revision: 1.29 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <math.h>
#include <cpl.h>

#include "irplib_oddeven.h"

#include "isaac_utils.h"
#include "isaac_pfits.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup isaac_utils     Miscellaneous Utilities
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the pipeline copyright and license
  @return   The copyright and license string

  The function returns a pointer to the statically allocated license string.
  This string should not be modified using the returned pointer.
 */
/*----------------------------------------------------------------------------*/
const char * isaac_get_license(void)
{
    const char  *   isaac_license = 
        "This file is part of the ISAAC Instrument Pipeline\n"
        "Copyright (C) 2002,2003 European Southern Observatory\n"
        "\n"
        "This program is free software; you can redistribute it and/or modify\n"
        "it under the terms of the GNU General Public License as published by\n"
        "the Free Software Foundation; either version 2 of the License, or\n"
        "(at your option) any later version.\n"
        "\n"
        "This program is distributed in the hope that it will be useful,\n"
        "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
        "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
        "GNU General Public License for more details.\n"
        "\n"
        "You should have received a copy of the GNU General Public License\n"
        "along with this program; if not, write to the Free Software\n"
        "Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, \n"
        "MA  02111-1307  USA";
    return isaac_license;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the offsets from a set of frames
  @param    fset    the set of frames
  @return   The offsets in a bivector or NULL in error case
 */
/*----------------------------------------------------------------------------*/
cpl_bivector * isaac_get_offsets(const cpl_frameset * fset)
{
    cpl_bivector        *   offsets;
    double              *   offsets_x;
    double              *   offsets_y;
    const cpl_frame     *   frame;
    cpl_propertylist    *   plist;
    int                     nfiles;
    int                     i;

    /* Test entries */
    if (fset == NULL) return NULL;

    /* Create the offsets bi vector */
    nfiles = cpl_frameset_get_size(fset);
    offsets = cpl_bivector_new(nfiles);
    offsets_x = cpl_bivector_get_x_data(offsets);
    offsets_y = cpl_bivector_get_y_data(offsets); 
    for (i=0; i<nfiles; i++) {
        if (cpl_error_get_code()) {
            cpl_bivector_delete(offsets);
            return NULL;
        }
        /* X and Y offsets */
        frame = cpl_frameset_get_position_const(fset, i);
        plist=cpl_propertylist_load(cpl_frame_get_filename(frame),0);
        offsets_x[i] = isaac_pfits_get_cumoffsetx(plist);
        offsets_y[i] = isaac_pfits_get_cumoffsety(plist);
        cpl_propertylist_delete(plist);
        if (cpl_error_get_code()) {
            cpl_msg_error(cpl_func, "Cannot get offsets from header");
            cpl_bivector_delete(offsets);
            return NULL;
        }
    }
    /* Subtract the first offset to all offsets */
    for (i=1; i<nfiles; i++) {
        offsets_x[i] -= offsets_x[0];
        offsets_y[i] -= offsets_y[0];
    }
    offsets_x[0] = offsets_y[0] = 0.00;

    return offsets;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the broad band filter
  @param    f           the filter name
  @return   the broadband filter id
 */
/*----------------------------------------------------------------------------*/
isaac_band isaac_get_bbfilter(const char * f)
{
    if (!strcmp(f, "J"))            return ISAAC_BAND_J;
    if (!strcmp(f, "J+Block"))      return ISAAC_BAND_J;
    if (!strcmp(f, "Js"))           return ISAAC_BAND_J;
    if (!strcmp(f, "Z"))            return ISAAC_BAND_J;
    if (!strcmp(f, "SZ"))           return ISAAC_BAND_J;
    if (!strcmp(f, "SH"))           return ISAAC_BAND_H;
    if (!strcmp(f, "H"))            return ISAAC_BAND_H;
    if (!strcmp(f, "Ks"))           return ISAAC_BAND_KS;
    if (!strcmp(f, "K"))            return ISAAC_BAND_K;
    if (!strcmp(f, "SK"))           return ISAAC_BAND_K;
    if (!strcmp(f, "L"))            return ISAAC_BAND_L;
    if (!strcmp(f, "L-new"))        return ISAAC_BAND_L;
    if (!strcmp(f, "SL"))           return ISAAC_BAND_L;
    if (!strcmp(f, "M"))            return ISAAC_BAND_M;
    if (!strcmp(f, "M_NB"))         return ISAAC_BAND_M;
    if (!strcmp(f, "M_NB-new"))     return ISAAC_BAND_M;
    if (!strcmp(f, "NB_1.06"))      return ISAAC_BAND_J;
    if (!strcmp(f, "NB_1.08"))      return ISAAC_BAND_J;
    if (!strcmp(f, "NB_1.19"))      return ISAAC_BAND_J;
    if (!strcmp(f, "NB_1.21"))      return ISAAC_BAND_J;
    if (!strcmp(f, "NB_1.26"))      return ISAAC_BAND_J;
    if (!strcmp(f, "NB_1.28"))      return ISAAC_BAND_J;
    if (!strcmp(f, "NB_1.64"))      return ISAAC_BAND_H;
    if (!strcmp(f, "NB_1.71"))      return ISAAC_BAND_H;
    if (!strcmp(f, "NB_2.07"))      return ISAAC_BAND_KS;
    if (!strcmp(f, "NB_2.09"))      return ISAAC_BAND_KS;
    if (!strcmp(f, "NB_2.13"))      return ISAAC_BAND_KS;
    if (!strcmp(f, "NB_2.17"))      return ISAAC_BAND_KS;
    if (!strcmp(f, "NB_2.19"))      return ISAAC_BAND_KS;
    if (!strcmp(f, "NB_2.25"))      return ISAAC_BAND_KS;
    if (!strcmp(f, "NB_2.29"))      return ISAAC_BAND_KS;
    if (!strcmp(f, "NB_2.34"))      return ISAAC_BAND_KS;
    if (!strcmp(f, "NB_3.21"))      return ISAAC_BAND_L;
    if (!strcmp(f, "NB_3.28"))      return ISAAC_BAND_L;
    if (!strcmp(f, "NB_3.80"))      return ISAAC_BAND_L;
    if (!strcmp(f, "NB_4.07"))      return ISAAC_BAND_L;

    (void)cpl_error_set_message(cpl_func, CPL_ERROR_UNSUPPORTED_MODE, "%s", f);

    return ISAAC_BAND_UNKNOWN;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the broad band filter
  @param    f           the filter name
  @return   the broadband filter id
 */
/*----------------------------------------------------------------------------*/
isaac_band isaac_get_associated_filter(const char * f)
{
    if (!strcmp(f, "J"))            return ISAAC_BAND_J;
    if (!strcmp(f, "Js"))           return ISAAC_BAND_JS;
    if (!strcmp(f, "Z"))            return ISAAC_BAND_Z;
    if (!strcmp(f, "SZ"))           return ISAAC_BAND_SZ;
    if (!strcmp(f, "SH"))           return ISAAC_BAND_SH;
    if (!strcmp(f, "H"))            return ISAAC_BAND_H;
    if (!strcmp(f, "Ks"))           return ISAAC_BAND_KS;
    if (!strcmp(f, "K"))            return ISAAC_BAND_K;
    if (!strcmp(f, "SK"))           return ISAAC_BAND_SK;
    if (!strcmp(f, "L"))            return ISAAC_BAND_L;
    if (!strcmp(f, "SL"))           return ISAAC_BAND_SL;
    if (!strcmp(f, "M"))            return ISAAC_BAND_M;
    if (!strcmp(f, "M_NB"))         return ISAAC_BAND_M;
    if (!strcmp(f, "NB_1.06"))      return ISAAC_BAND_J;
    if (!strcmp(f, "NB_1.08"))      return ISAAC_BAND_J;
    if (!strcmp(f, "NB_1.19"))      return ISAAC_BAND_J;
    if (!strcmp(f, "NB_1.21"))      return ISAAC_BAND_J;
    if (!strcmp(f, "NB_1.26"))      return ISAAC_BAND_J;
    if (!strcmp(f, "NB_1.28"))      return ISAAC_BAND_J;
    if (!strcmp(f, "NB_1.64"))      return ISAAC_BAND_H;
    if (!strcmp(f, "NB_1.71"))      return ISAAC_BAND_H;
    if (!strcmp(f, "NB_2.07"))      return ISAAC_BAND_KS;
    if (!strcmp(f, "NB_2.09"))      return ISAAC_BAND_KS;
    if (!strcmp(f, "NB_2.13"))      return ISAAC_BAND_KS;
    if (!strcmp(f, "NB_2.17"))      return ISAAC_BAND_KS;
    if (!strcmp(f, "NB_2.19"))      return ISAAC_BAND_KS;
    if (!strcmp(f, "NB_2.25"))      return ISAAC_BAND_KS;
    if (!strcmp(f, "NB_2.29"))      return ISAAC_BAND_KS;
    if (!strcmp(f, "NB_2.34"))      return ISAAC_BAND_KS;
    if (!strcmp(f, "NB_3.21"))      return ISAAC_BAND_L;
    if (!strcmp(f, "NB_3.28"))      return ISAAC_BAND_L;
    if (!strcmp(f, "NB_3.80"))      return ISAAC_BAND_L;
    if (!strcmp(f, "NB_4.07"))      return ISAAC_BAND_L;

    (void)cpl_error_set_message(cpl_func, CPL_ERROR_UNSUPPORTED_MODE, "%s", f);

    return ISAAC_BAND_UNKNOWN;
}

/*-------------------------------------------------------------------------*/
/**
  @brief    Return a band name
  @param    band    a BB
  @return   1 pointer to a static band name.
 */
/*--------------------------------------------------------------------------*/
const char * isaac_std_band_name(isaac_band band)
{
    switch (band) {
        case ISAAC_BAND_J:        return "J";
        case ISAAC_BAND_JS:       return "Js";
        case ISAAC_BAND_JBLOCK:   return "J+Block";
        case ISAAC_BAND_H:        return "H";
        case ISAAC_BAND_K:        return "K";
        case ISAAC_BAND_KS:       return "Ks";
        case ISAAC_BAND_L:        return "L";
        case ISAAC_BAND_M:        return "M";
        case ISAAC_BAND_LP:       return "Lp";
        case ISAAC_BAND_MP:       return "Mp";
        case ISAAC_BAND_Z:        return "Z";
        case ISAAC_BAND_SZ:       return "SZ";
        case ISAAC_BAND_SH:       return "SH";
        case ISAAC_BAND_SK:       return "SK";
        case ISAAC_BAND_SL:       return "SL";
        default:
            (void)cpl_error_set_message(cpl_func, CPL_ERROR_UNSUPPORTED_MODE,
                                        "%d", (int)band);
            return "Unknown";
    } 
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Correct the odd/even in an image 
  @param    in      the inpute image
  @return   the corrected image or NULL on error case
*/
/*----------------------------------------------------------------------------*/
cpl_image * isaac_oddeven_correct(const cpl_image * in)
{
    cpl_image       *   quad1;
    cpl_image       *   quad2;
    cpl_image       *   quad3;
    cpl_image       *   quad4;
    cpl_image       *   quad1_corr;
    cpl_image       *   quad2_corr;
    cpl_image       *   quad3_corr;
    cpl_image       *   quad4_corr;
    float           *   pout;
    float           *   pin1;
    float           *   pin2;
    float           *   pin3;
    float           *   pin4;
    int                 nx, ny;
    cpl_image       *   out;
    int                 i, j, pos1, pos2, pos3, pos4;

    /* Test entries */
    if (in == NULL) return NULL;

    /* Initialise */
    nx = cpl_image_get_size_x(in);
    ny = cpl_image_get_size_y(in);

    /* Extract quadrants */
    quad1 = cpl_image_extract(in, 1, 1, nx/2, ny/2);
    quad2 = cpl_image_extract(in, (nx/2)+1, 1, nx, ny/2);
    quad3 = cpl_image_extract(in, 1, (ny/2)+1, nx/2, ny);
    quad4 = cpl_image_extract(in, (nx/2)+1, (ny/2)+1, nx, ny);
    
    /* Apply the FFT to the quadrants */
    quad1_corr = irplib_oddeven_correct(quad1);
    quad2_corr = irplib_oddeven_correct(quad2);
    quad3_corr = irplib_oddeven_correct(quad3);
    quad4_corr = irplib_oddeven_correct(quad4);

    cpl_image_delete(quad1);
    cpl_image_delete(quad2);
    cpl_image_delete(quad3);
    cpl_image_delete(quad4);
    
    /* Construct the output image */
    out = cpl_image_duplicate(in);
    pout = cpl_image_get_data_float(out);
    pin1 = cpl_image_get_data_float(quad1_corr);
    pin2 = cpl_image_get_data_float(quad2_corr);
    pin3 = cpl_image_get_data_float(quad3_corr);
    pin4 = cpl_image_get_data_float(quad4_corr);

    for (j=0; j<ny/2; j++) {
        for (i=0; i<nx/2; i++) {
            pos1 = i + j * nx;
            pos2 = (i+(nx/2)) + j * nx;
            pos3 = i + (j+(ny/2)) * nx;
            pos4 = (i+(nx/2)) + (j+(ny/2)) * nx;
            pout[pos1] = pin1[i+j*(nx/2)];
            pout[pos2] = pin2[i+j*(nx/2)];
            pout[pos3] = pin3[i+j*(nx/2)];
            pout[pos4] = pin4[i+j*(nx/2)];
        }
    }
    
    cpl_image_delete(quad1_corr);
    cpl_image_delete(quad2_corr);
    cpl_image_delete(quad3_corr);
    cpl_image_delete(quad4_corr);
    return out;
}

/*----------------------------------------------------------------------------*/
/**
   @brief   Extract the frames with the given tag from a frameset
   @param   self    A frameset
   @param   tag     The tag of the requested frames   
   @return  The newly created frameset or NULL when none found, or on error

   The returned frameset must be de allocated with cpl_frameset_delete()
 */
/*----------------------------------------------------------------------------*/
cpl_frameset * isaac_extract_frameset(const cpl_frameset * self,
                                      const char         * tag)
{
    const cpl_size nframes = cpl_frameset_get_size(self);
    cpl_frameset    * newset = NULL;


    if (self == NULL) return NULL;
    if (tag  == NULL) return NULL;

    for (cpl_size i = 0; i < nframes; i++) {
        const cpl_frame * frame = cpl_frameset_get_position_const(self, i);
        const char * mytag = cpl_frame_get_tag(frame);

        if (mytag != NULL && !strcmp(mytag, tag)) {
            cpl_frame * newframe = cpl_frame_duplicate(frame);

            /* Create the output frameset */
            if (newset == NULL) newset = cpl_frameset_new();

            if (cpl_frameset_insert(newset, newframe)) {
                cpl_frame_delete(newframe);
                cpl_frameset_delete(newset);
                (void)cpl_error_set(cpl_func, cpl_error_get_code());
                return NULL;
            }
        }
    }

    return newset;
}

/*----------------------------------------------------------------------------*/
/** 
   @brief   Extract the filename of the first frame of the given tag
   @param   self    A frameset
   @param   tag     The tag of the requested file 
   @return  Pointer to the file, or NULL when not found
   @note The pointer is only valid as long as the frame exists
 */
/*----------------------------------------------------------------------------*/
const char * isaac_extract_filename(const cpl_frameset * self,
                                    const char         * tag)
{
    const cpl_frame * frame;

    if (self == NULL) return NULL;
    if (tag  == NULL) return NULL;

    /* Get the frame  */
    frame = cpl_frameset_find_const(self, tag);

    return frame == NULL ? NULL : cpl_frame_get_filename(frame);
}

/**@}*/
