/* $Id: isaac_img_slitpos.c,v 1.36 2013-03-12 08:06:48 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:06:48 $
 * $Revision: 1.36 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <math.h>
#include <cpl.h>

#include "irplib_utils.h"
#include "irplib_slitpos.h"

#include "isaac_utils.h"
#include "isaac_pfits.h"
#include "isaac_dfs.h"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int isaac_img_slitpos_create(cpl_plugin *);
static int isaac_img_slitpos_exec(cpl_plugin *);
static int isaac_img_slitpos_destroy(cpl_plugin *);
static int isaac_img_slitpos(cpl_parameterlist *, cpl_frameset *);
static int isaac_img_slitpos_save(cpl_table *, int, cpl_parameterlist *, 
        cpl_frameset *);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    double      slit_width;
    int         products_flag;
    /* Outputs */
    double      angle;
    int         slit_length;
    double      xpos;
    double      ypos;
    double      slit_flux;
} isaac_img_slitpos_config;

static char isaac_img_slitpos_description[] = 
"isaac_img_slitpos -- ISAAC imaging slit position recipe.\n"
"The files listed in the Set Of Frames (sof-file) must be tagged:\n"
"raw-file.fits " ISAAC_IMG_SLITPOS_RAW " or"
"raw-file.fits " ISAAC_IMG_SLITPOS_CAL_RAW "\n";

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe));
    cpl_plugin  *   plugin = &recipe->interface;

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    ISAAC_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "isaac_img_slitpos",
                    "Slit position recipe",
                    isaac_img_slitpos_description,
                    "Lars Lundin",
                    PACKAGE_BUGREPORT,
                    isaac_get_license(),
                    isaac_img_slitpos_create,
                    isaac_img_slitpos_exec,
                    isaac_img_slitpos_destroy);

    cpl_pluginlist_append(list, plugin);
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using the 
  interface. 
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_slitpos_create(cpl_plugin * plugin)
{
    cpl_recipe      * recipe;
    cpl_parameter   * p;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new();

    /* Fill the parameters list */
    /* --slit_w */
    p = cpl_parameter_new_value("isaac.isaac_img_slitpos.slit_width",
            CPL_TYPE_DOUBLE, "Slit width", "isaac.isaac_img_slitpos",
            20.0);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "slit_w");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* --prod */
    p = cpl_parameter_new_value("isaac.isaac_img_slitpos.products",
            CPL_TYPE_BOOL, "flag to create the products", 
            "isaac.isaac_img_slitpos", TRUE);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "prod");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);

    /* Return */
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_slitpos_exec(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    return isaac_img_slitpos(recipe->parameters, recipe->frames);
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_slitpos_destroy(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    cpl_parameterlist_delete(recipe->parameters);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_slitpos(
        cpl_parameterlist   *   parlist, 
        cpl_frameset        *   framelist)
{
    cpl_parameter       *   par;
    cpl_imagelist       *   imlist;
    cpl_propertylist    *   plist;
    cpl_frame           *   cur_frame;
    cpl_frameset        *   slitframes;
    const char          *   sval;
    cpl_table           *   out_table;
    int                     slit_length;
    int                     with_dark;
    int                     slit_ind;
    cpl_image           *   dark;
    double                  loc_pi;
    cpl_boolean             did_reduce = CPL_FALSE;
    int                     i;

    /* Initialise */
    isaac_img_slitpos_config.angle = -1.0;
    isaac_img_slitpos_config.slit_flux = 0.0;
    isaac_img_slitpos_config.xpos = -1.0;
    isaac_img_slitpos_config.ypos = -1.0;
    loc_pi = 3.1415926535897932384626433;
    
    /* Retrieve input parameters */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_img_slitpos.slit_width");
    isaac_img_slitpos_config.slit_width = cpl_parameter_get_double(par);
    par = cpl_parameterlist_find(parlist, "isaac.isaac_img_slitpos.products"); 
    isaac_img_slitpos_config.products_flag = cpl_parameter_get_bool(par);

    /* Identify the RAW and CALIB frames in the input frameset */
    if (isaac_dfs_set_groups(framelist)) {
        cpl_msg_error(cpl_func, "Cannot identify RAW and CALIB frames");
        return CPL_ERROR_UNSPECIFIED;
    }

    slitframes = isaac_extract_frameset(framelist, ISAAC_IMG_SLITPOS_RAW);
    if (slitframes == NULL) 
        slitframes = isaac_extract_frameset(framelist,
                                            ISAAC_IMG_SLITPOS_CAL_RAW);
    cpl_ensure_code(slitframes != NULL, CPL_ERROR_DATA_NOT_FOUND);

    /* Load the data */
    imlist = cpl_imagelist_load_frameset(slitframes, CPL_TYPE_FLOAT, 1, 0);
    cpl_frameset_delete(slitframes);
    
    /* Check if the dark frame is there */
    /* Get FITS header from reference file */
    cur_frame = cpl_frameset_get_position(framelist, 0);
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(cur_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from first frame");
        cpl_imagelist_delete(imlist);
        return CPL_ERROR_UNSPECIFIED;
    }
    if ((sval = isaac_pfits_get_mode(plist)) == NULL) {
        cpl_msg_error(cpl_func, "getting the MODE");
        cpl_propertylist_delete(plist);
        cpl_imagelist_delete(imlist);
        return CPL_ERROR_UNSPECIFIED;
    }
    with_dark = 0;
    if (!strcmp(sval, "SW_DARK") || !strcmp(sval, "LW_DARK")) {
        dark = cpl_image_duplicate(cpl_imagelist_get(imlist, 0));
        cpl_imagelist_subtract_image(imlist, dark);
        cpl_image_delete(dark);
        with_dark = 1;
    }
    cpl_propertylist_delete(plist);
    
    /* Reduce each frame separately */
    for (i=0; i<cpl_imagelist_get_size(imlist); i++) {
        cpl_errorstate prestate = cpl_errorstate_get();
        if(i==0 && with_dark) continue;
        if (with_dark) slit_ind = i-1;
        else slit_ind = i;
        /* Slit analysis */
        cpl_msg_info(cpl_func, "Analyse slit nb %d", slit_ind+1);
        cpl_msg_indent_more();
        if ((out_table = irplib_slitpos_analysis(
                        cpl_imagelist_get(imlist, i),
                        isaac_img_slitpos_config.slit_width, 
                        &(isaac_img_slitpos_config.slit_flux))) == NULL) {
            cpl_msg_error(cpl_func, "Could not analyse the slit in image "
                          "%d/%d", i+1, (int)cpl_imagelist_get_size(imlist));
            cpl_errorstate_dump(prestate, CPL_FALSE, NULL);
            cpl_errorstate_set(prestate);
            cpl_msg_indent_less();
            continue;
        }
        
        /* Find the slit angle in degrees with the horizontal axis   */
        slit_length = cpl_table_get_nrow(out_table);
        /* THE ANGLE */
        isaac_img_slitpos_config.angle = (double)atan(
            (double)(cpl_table_get_int(out_table,"SLIT_Y",slit_length-1, NULL) -
                     cpl_table_get_int(out_table, "SLIT_Y", 0, NULL)) /
            (cpl_table_get_double(out_table,"SLIT_CENTER",slit_length-1, NULL) -
             cpl_table_get_double(out_table, "SLIT_CENTER", 0, NULL)));
        isaac_img_slitpos_config.angle *= 180.0/loc_pi;
        if (isaac_img_slitpos_config.angle >= 0.00 && 
                isaac_img_slitpos_config.angle < 90.0)
            isaac_img_slitpos_config.angle += 180.0;
        /* THE SLIT LENGTH */
        isaac_img_slitpos_config.slit_length = slit_length;
        /* THE SLIT CENTER POSITION */
        isaac_img_slitpos_config.xpos = 
            (cpl_table_get_double(out_table,"SLIT_CENTER",0,NULL) +
             cpl_table_get_double(out_table,"SLIT_CENTER",slit_length-1,NULL))
            / 2.0;
        isaac_img_slitpos_config.ypos = 
            (double)(cpl_table_get_int(out_table,"SLIT_Y",0,NULL) +
             cpl_table_get_int(out_table,"SLIT_Y",slit_length-1,NULL))
            / 2.0;
        
        /* Print the results */
        cpl_msg_info(cpl_func, "Slit flux   : %g", 
                isaac_img_slitpos_config.slit_flux);
        cpl_msg_info(cpl_func, "Slit angle  : %g", 
                isaac_img_slitpos_config.angle);
        cpl_msg_info(cpl_func, "Slit length : %d", 
                isaac_img_slitpos_config.slit_length);
        cpl_msg_info(cpl_func, "Slit center : %g %g", 
                isaac_img_slitpos_config.xpos,
                isaac_img_slitpos_config.ypos);
        cpl_msg_indent_less();

        /* Save the product */
        if (isaac_img_slitpos_config.products_flag) {
            cpl_msg_info(cpl_func, "Save the product");
            cpl_msg_indent_more();
            if (isaac_img_slitpos_save(out_table, slit_ind+1, parlist, 
                        framelist) == -1) {
                cpl_msg_error(cpl_func, "Cannot save the products");
                cpl_imagelist_delete(imlist);
                cpl_table_delete(out_table);
                cpl_msg_indent_less();
                return CPL_ERROR_UNSPECIFIED;
            }
            cpl_msg_indent_less();
        }
        cpl_table_delete(out_table);
        if (!cpl_error_get_code()) did_reduce = CPL_TRUE;
    }
            
    /* Free and return */
    cpl_imagelist_delete(imlist);

    cpl_ensure_code(did_reduce, CPL_ERROR_ILLEGAL_INPUT);

    return cpl_error_set_where(cpl_func); /* Propagate error, if any */
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the product of the recipe
  @param    out_table   the table with the slit position
  @param    file_nb     file number
  @param    parlist     the input list of parameters
  @param    set         the input frame set
  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_slitpos_save(
        cpl_table           *   out_table,   
        int                     file_nb,
        cpl_parameterlist   *   parlist,
        cpl_frameset        *   set)
{
    cpl_propertylist * plist;
    cpl_propertylist * paflist;
    cpl_propertylist * qclist;
    /* Get the reference frame */
    const cpl_frame  * ref_frame = cpl_frameset_get_position_const(set, file_nb-1);
    char             * filename;

    /* Get the QC params in qclist */
    qclist = cpl_propertylist_new();
    cpl_propertylist_append_double(qclist, "ESO QC SLIT FLUX",
            isaac_img_slitpos_config.slit_flux);
    cpl_propertylist_append_double(qclist, "ESO QC SLIT POSANG",
            isaac_img_slitpos_config.angle);
    cpl_propertylist_append_double(qclist, "ESO QC SLIT XPOS",
            isaac_img_slitpos_config.xpos);
    cpl_propertylist_append_double(qclist, "ESO QC SLIT YPOS",
            isaac_img_slitpos_config.ypos);

    cpl_propertylist_append_string(qclist, CPL_DFS_PRO_CATG,
                                   ISAAC_IMG_SLITPOS_RES);

    /* Write the table */
    filename = cpl_sprintf("isaac_img_slitpos_%02d.fits", file_nb);

    cpl_dfs_save_table(set, NULL, parlist, set, ref_frame, out_table,
                       NULL, "isaac_img_slitpos", qclist, NULL,
                       PACKAGE "/" PACKAGE_VERSION, filename);
    cpl_free(filename);

    /* Get FITS header from reference file */
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Get the keywords for the paf file */
    paflist = cpl_propertylist_new();
    cpl_propertylist_copy_property_regexp(paflist, plist,
        "^(ARCFILE|ORIGFILE|MJD-OBS|INSTRUME|ESO TPL ID|ESO TPL NEXP|"
        "ESO TPL EXPNO|ESO DPR CATG|ESO DPR TECH|ESO DPR TYPE|DATE-OBS|"
        "ESO INS OPTI1 ID)$",0);
    cpl_propertylist_delete(plist);

    /* Copy the QC in paflist */
    cpl_propertylist_copy_property_regexp(paflist, qclist, ".", 0);
    cpl_propertylist_delete(qclist);

    /* PRO.CATG */
    cpl_propertylist_update_string(paflist, CPL_DFS_PRO_CATG,
                                   ISAAC_IMG_SLITPOS_RES);

    /* Save the PAF file */
    filename = cpl_sprintf("isaac_img_slitpos_%02d.paf", file_nb);
    cpl_dfs_save_paf("ISAAC",
            "isaac_img_slitpos",
            paflist,
            filename);
    cpl_free(filename);
    cpl_propertylist_delete(paflist);
    return  0;
}

